// Desktop Rox - A game by Stephen Stair (sgstair@akkit.org) for LudumDare 11 (48h game competition)
// DesktopSpriteManager - system for rendering sprites on the desktop :>

#include "DesktopSpriteManager.h"

#include "BitmapSprite.h"

#include <vector>
#include <algorithm>
using namespace std;

extern HINSTANCE g_hInst;






vector< pair<int,int> > pointlist;
vector< pair<int,int> > linelist;

int AddPoint(int x, int y)
{
	vector< pair<int,int> >::iterator i;
	i=find(pointlist.begin(),pointlist.end(),pair<int,int>(x,y));
	if(i!=pointlist.end()) return (int)(i-pointlist.begin());
	pointlist.push_back( pair<int,int>(x,y) );
	return (int) pointlist.size()-1;
}
int GetPointX(int i)
{
	return pointlist[i].first;
}
int GetPointY(int i)
{
	return pointlist[i].second;
}
int FindLineEnd(int i)
{
	for(size_t n=0; n<linelist.size(); n++)
	{
		if(linelist[n].first == i) return linelist[n].second;
	}
	return -1;
}

bool lesspair(const pair<int,int> & a, const pair<int,int> & b)
{
	return a.first < b.first;
}
void AddLine(int x1, int y1, int x2, int y2)
{
	// find all of the points along this line, and add a segment between each 2.
	vector< pair<int,int> > stops;
	size_t i;
	int mn,mx;
	bool reverse = false;
	if(x1==x2)
	{ // vertical!
		if(y1<=y2) { mn=y1; mx=y2; } else { mn=y2; mx=y1; reverse=true;}
		for(i=0;i<pointlist.size();i++)
		{
			if(pointlist[i].first==x1)
			{
				if(pointlist[i].second>=mn && pointlist[i].second<=mx)
				{
					stops.push_back( pair<int,int>( pointlist[i].second, (int)i) );
				}
			}
		}
	}
	else
	{ // horizontal!
		if(x1<=x2) { mn=x1; mx=x2; } else { mn=x2; mx=x1; reverse = true; }
		for(i=0;i<pointlist.size();i++)
		{
			if(pointlist[i].second==y1)
			{
				if(pointlist[i].first>=mn && pointlist[i].first<=mx)
				{
					stops.push_back( pair<int,int>( pointlist[i].first, (int)i) );
				}
			}
		}
	}
	if(reverse)
	{
		sort(stops.rbegin(), stops.rend(), lesspair);
	}
	else
	{
		sort(stops.begin(), stops.end(), lesspair);
	}
	for(i=1;i<stops.size();i++)
	{
		linelist.push_back( pair<int,int>( stops[i-1].second, stops[i].second ) );
	}
}

vector<BoundaryPoint> CalcBoundary( vector<RECT> & windows )
{
	pointlist.clear();
	linelist.clear();
	pointlist.reserve(windows.size()*4);
	linelist.reserve( windows.size()*4);

	size_t i,j;
	// add points
	for(i=0;i<windows.size();i++)
	{
		AddPoint(windows[i].left, windows[i].top);
		AddPoint(windows[i].right, windows[i].top);
		AddPoint(windows[i].right, windows[i].bottom);
		AddPoint(windows[i].left, windows[i].bottom);
	}
	// add lines
	for(i=0;i<windows.size();i++)
	{
		AddLine(windows[i].left, windows[i].top,windows[i].right, windows[i].top);
		AddLine(windows[i].right, windows[i].top,windows[i].right, windows[i].bottom);
		AddLine(windows[i].right, windows[i].bottom,windows[i].left, windows[i].bottom);
		AddLine(windows[i].left, windows[i].bottom,windows[i].left, windows[i].top);
	}

	// remove duplicate lines
	pair<int,int> tp;
	for(i=0;i<linelist.size();i++)
	{
		tp = linelist[i];
		for(j=i+1;j<linelist.size();j++)
		{
			if(linelist[j]!=tp)
			{
				if(linelist[j].first!=tp.second || linelist[j].second != tp.first)
				{
					continue;
				}
			} // lines are the same; delete!
			linelist.erase(linelist.begin()+j);
			linelist.erase(linelist.begin()+i);
			i--;
			break;
		}
	}

	// trace out boundary
	vector<BoundaryPoint> temp;
	int first_id = linelist[0].first;
	int id=first_id;
	while(id!=-1) {
		temp.push_back(BoundaryPoint());
		temp.back().x = GetPointX(id);
		temp.back().y = GetPointY(id);
		id = FindLineEnd(id);
		if(id==first_id) break; // done!
	}
	pointlist.clear();
	linelist.clear();
	return temp;
}

















int DesktopSpriteManager::x_offset=0, DesktopSpriteManager::y_offset=0;

class StupidLine
{
public:
	StupidLine() {x1=y1=x2=y2=0;}
	StupidLine( int x1_, int y1_, int x2_, int y2_) { x1=x1_; y1=y1_; x2=x2_; y2=y2_; }
	int x1, y1, x2, y2;
};


void DesktopSpriteManager::UpdateDSRWindow(DesktopSpriteRecord * dsr)
{
	dsr->cachedbmp.resize( dsr->sprite->bmp.wid(), dsr->sprite->bmp.hei() );
	memcpy( dsr->cachedbmp.bits(), dsr->sprite->bmp.bits(), dsr->sprite->bmp.wid()*dsr->sprite->bmp.hei()*4 );
	//dsr->cachedbmp.MakeLayeredAlpha();

	MoveWindow(dsr->hWnd, dsr->sprite->bx+x_offset, dsr->sprite->by+y_offset, dsr->cachedbmp.wid(), dsr->cachedbmp.hei(), false);

	//GetClientRect(dsr->hWnd,&rc);
	//InvalidateRect(dsr->hWnd,&rc,false);
	{
		POINT pos, src;
		SIZE dims;
		pos.x = dsr->sprite->bx+x_offset;
		pos.y = dsr->sprite->by+y_offset;
		src.x=src.y=0;
		dims.cx=dsr->cachedbmp.wid();
		dims.cy=dsr->cachedbmp.hei();

		BLENDFUNCTION bf;
		bf.BlendOp=AC_SRC_OVER;
		bf.BlendFlags=0;
		bf.SourceConstantAlpha=255;
		bf.AlphaFormat=AC_SRC_ALPHA;
		UpdateLayeredWindow(dsr->hWnd,0,&pos,&dims,dsr->cachedbmp.DC(),&src,0,&bf,ULW_ALPHA);
	}

}


BOOL CALLBACK DesktopSpriteManager::EnumProc( HMONITOR hMon, HDC hDC, LPRECT lpmonitorrect, LPARAM data )
{
	DesktopSpriteManager * dsm = (DesktopSpriteManager *) data;
	dsm->windows.push_back(*lpmonitorrect);
	return TRUE;
}

DesktopSpriteManager::DesktopSpriteManager()
{
	WNDCLASS wc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hbrBackground=0;
	wc.hCursor=LoadCursor(NULL,MAKEINTRESOURCE(IDC_ARROW));
	wc.hIcon=0;
	wc.hInstance=g_hInst;
	wc.lpfnWndProc=DesktopSpriteManager::WndProc;
	wc.lpszClassName="DSM_class";
	wc.lpszMenuName=0;
	wc.style=0;
	RegisterClass(&wc);


	primarywindow=0;
	x_offset = 0;
	y_offset = 0;
	width = 0;
	height = 0;
	//vector<RECT> windows;
	//int x_offset, y_offset;
	//int width, height;
	//vector<BoundaryPoint> cachedboundary;

	// figure out how big the screen is, and store the data!
	EnumDisplayMonitors(NULL,NULL, EnumProc, (LPARAM)this);
	if(windows.empty())
	{
		RECT rc; // shovel something in it.
		rc.top=0;
		rc.left=0;
		rc.bottom=768;
		rc.right=1024;
		windows.push_back(rc);
	}
	int min_x, max_x, min_y, max_y;
	min_x = max_x = min_y = max_y = 0;
	for(size_t i=0;i<windows.size();i++)
	{
		if(windows[i].top==0 && windows[i].left==0) primarywindow=(int)i;
		if(windows[i].top<min_y)  min_y = windows[i].top;
		if(windows[i].left<min_x)  min_x = windows[i].left;
		if(windows[i].right>max_x)  max_x = windows[i].right;
		if(windows[i].bottom>max_y)  max_y = windows[i].bottom;
	}
	width = max_x - min_x;
	height = max_y - min_y;
	x_offset = min_x;
	y_offset = min_y;
	// adjust window rects for our coordinate system
	for(size_t i=0;i<windows.size();i++)
	{
		windows[i].top -= y_offset;
		windows[i].left -= x_offset;
		windows[i].right -= x_offset;
		windows[i].bottom -= y_offset;
	}
	// find boundary -_-
	cachedboundary = CalcBoundary(windows);
}
DesktopSpriteManager::~DesktopSpriteManager(void)
{
	size_t i;
	for(i=0;i<mysprites.size();i++)
	{
		SetWindowLong(mysprites[i]->hWnd, GWL_USERDATA,0);
		DestroyWindow( mysprites[i]->hWnd );
		delete mysprites[i];
	}
	mysprites.clear();
}



int DesktopSpriteManager::GetFieldWidth()
{
	return width;
}
int DesktopSpriteManager::GetFieldHeight()
{
	return height;
}
vector<BoundaryPoint> DesktopSpriteManager::GetFieldBoundary()
{ 
	return cachedboundary;
}
RECT DesktopSpriteManager::GetMainWindow()
{
	return windows[primarywindow];
}

void DesktopSpriteManager::SetClickHandler(fptrClickHandler h)
{
	myhandler = h;
}

void DesktopSpriteManager::AddSprite( BitmapSprite * sprite )
{
	RemoveSprite(sprite); // precautionary
	DesktopSpriteRecord * dsr = new DesktopSpriteRecord();
	dsr->sprite = sprite;
	dsr->hWnd = CreateWindowEx(WS_EX_LAYERED|WS_EX_TOPMOST|WS_EX_TOOLWINDOW,"DSM_class","dsm_window",0,0,0,sprite->bmp.wid(),sprite->bmp.hei(),0,0,g_hInst,dsr);
	SetWindowLongPtr(dsr->hWnd,GWL_USERDATA,(LONG)dsr);
	mysprites.push_back(dsr);
	UpdateDSRWindow(dsr);
	ShowWindow(dsr->hWnd,SW_SHOW);
}
bool DesktopSpriteManager::RemoveSprite( BitmapSprite * sprite )
{
	size_t i;
	for(i=0;i<mysprites.size();i++)
	{
		if(mysprites[i]->sprite==sprite)
		{
			DestroyWindow( mysprites[i]->hWnd );
			delete mysprites[i];
			mysprites.erase( mysprites.begin() + i );
			return true;
		}
	}
	return false;
}




void DesktopSpriteManager::Redraw()
{
	RECT rc;
	size_t i;
	DesktopSpriteRecord * dsr;
	for(i=0;i<mysprites.size();i++)
	{
		dsr = mysprites[i];
		if(dsr->sprite->alpha==0)
		{
			ShowWindow(dsr->hWnd,SW_HIDE);

		} else {
			ShowWindow(dsr->hWnd,SW_SHOW);
			UpdateDSRWindow(dsr);
			if(dsr->sprite->flags&FLAGS_SETFOCUS) if(GetFocus()) SetFocus(dsr->hWnd);
		}
	}
}

fptrClickHandler DesktopSpriteManager::myhandler = 0;

LRESULT WINAPI DesktopSpriteManager::WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	DesktopSpriteRecord * dsr = (DesktopSpriteRecord *)GetWindowLongPtr(hWnd,GWL_USERDATA);
	switch(msg)
	{
	case WM_CREATE:
		return 0;
	case WM_DESTROY:
		return 0;
	case WM_CLOSE:
		PostQuitMessage(0);
		return 0;

	case WM_RBUTTONDOWN:
		//PostQuitMessage(0);
		return 0;
	case WM_LBUTTONDOWN:
		{
			int x = LOWORD(lParam);
			int y = HIWORD(lParam);
			if(dsr && myhandler)
			{
				if(dsr->sprite->flags&FLAGS_WANTCLICKS)
				{
					myhandler(x,y,dsr->sprite);
				}
			}
		}
		return 0;

	case WM_PAINT:
		if(dsr)
		{
			UpdateDSRWindow(dsr);
		}
		else
		{
			PAINTSTRUCT ps;
			BeginPaint(hWnd,&ps);
			EndPaint(hWnd,&ps);
		}
		return 0;
	}
	return DefWindowProc(hWnd,msg,wParam,lParam);
}