// Desktop Rox - A game by Stephen Stair (sgstair@akkit.org) for LudumDare 11 (48h game competition)
// OriginalRasterizer - The first rasterizer idea I came up with. I'll probably add others too.

#include <math.h>
#include "OriginalRasterizer.h"

#include <vector>
using namespace std;

#define MY_PI 3.1415926535897932384626433832795
#define MY_SQR2 1.4142135623730950488016887242097

extern double line_thickness;
extern double shadow_thickness;


const Color foregroundcolor(255,255,255);
const Color backgroundcolor(0,0,0);

OriginalRasterizer::OriginalRasterizer(void)
{
	bound_bmp=0;
}

OriginalRasterizer::~OriginalRasterizer(void)
{
	bound_bmp=0;
}


void OriginalRasterizer::Bind(Bitmap * bmp, double zerox, double zeroy)
{
	bound_bmp = bmp;
	zx=zerox;
	zy=zeroy;
	if(bmp) bmp->cls();
}

void OriginalRasterizer::DrawClosedFigure( int num_points, RenderPoint * rp, double rotation, double scale )
{
	if(!bound_bmp || !rp) return;
	RenderPoint op[4];
	int i,i2;
	double nx,ny,l,dx,dy,q,px,py,sn,cs;
	// figure out normals!
	vector<double> pnx,pny, lnx,lny; // point normals, line normals
	pnx.resize(num_points,0.0);
	pny.resize(num_points,0.0);
	lnx.resize(num_points,0.0);
	lny.resize(num_points,0.0);
	for(i=0;i<num_points;i++)
	{
		nx = rp[(i+1)%num_points].y - rp[i].y;
		ny = -rp[(i+1)%num_points].x + rp[i].x;
		l=sqrt(nx*nx+ny*ny);
		if(l>0) { nx/=l; ny/=l; }
		lnx[i] = nx;
		lny[i] = ny;
		pnx[i] += nx;
		pny[i] += ny;
		pnx[(i+1)%num_points] += nx;
		pny[(i+1)%num_points] += ny;
	}
	// normalize point normals
	for(i=0;i<num_points;i++)
	{
		l=sqrt(pnx[i]*pnx[i] + pny[i]*pny[i]);
		if(l>0) {  pnx[i]/=l; pny[i]/=l; }
	}

	// draw shadow quads
	cs = cos( rotation );
	sn = sin( rotation );
	for(i=0;i<num_points;i++)
	{
		i2=(i+1)%num_points;
		l = (line_thickness+shadow_thickness*2)/2;
		q = lnx[i]*pnx[i] + lny[i]*pny[i]; // determine attenuation factor
		if(q==0) continue; // terminates in a straight line?
		l = l/q; // q is in range (0,1] where 0 is a 180 degree turn, and 1 is a 0 degree turn (straight line)
		dx = pnx[i]*l;
		dy = pny[i]*l;
		px = rp[i].x*scale + dx;
		py = rp[i].y*scale + dy;
		op[0].x = -zx + cs*px - sn*py;
		op[0].y = -zy + sn*px + cs*py;
		px = rp[i].x*scale - dx;
		py = rp[i].y*scale - dy;
		op[1].x = -zx + cs*px - sn*py;
		op[1].y = -zy + sn*px + cs*py;

		l = (line_thickness+shadow_thickness*2)/2; // same thing for other end of line!
		q = lnx[i]*pnx[i2] + lny[i]*pny[i2]; // determine attenuation factor
		if(q==0) continue; // terminates in a straight line?
		l = l/q; // q is in range (0,1] where 0 is a 180 degree turn, and 1 is a 0 degree turn (straight line)
		dx = pnx[i2]*l;
		dy = pny[i2]*l;
		px = rp[i2].x*scale - dx;
		py = rp[i2].y*scale - dy;
		op[2].x = -zx + cs*px - sn*py;
		op[2].y = -zy + sn*px + cs*py;
		px = rp[i2].x*scale + dx;
		py = rp[i2].y*scale + dy;
		op[3].x = -zx + cs*px - sn*py;
		op[3].y = -zy + sn*px + cs*py;

		DrawFilledQuad(op, backgroundcolor);
	}

	// draw line quads
	for(i=0;i<num_points;i++)
	{
		i2=(i+1)%num_points;
		l = (line_thickness)/2;
		q = lnx[i]*pnx[i] + lny[i]*pny[i]; // determine attenuation factor
		if(q==0) continue; // terminates in a straight line?
		l = l/q; // q is in range (0,1] where 0 is a 180 degree turn, and 1 is a 0 degree turn (straight line)
		dx = pnx[i]*l;
		dy = pny[i]*l;
		px = rp[i].x*scale + dx;
		py = rp[i].y*scale + dy;
		op[0].x = -zx + cs*px - sn*py;
		op[0].y = -zy + sn*px + cs*py;
		px = rp[i].x*scale - dx;
		py = rp[i].y*scale - dy;
		op[1].x = -zx + cs*px - sn*py;
		op[1].y = -zy + sn*px + cs*py;

		l = (line_thickness)/2; // same thing for other end of line!
		q = lnx[i]*pnx[i2] + lny[i]*pny[i2]; // determine attenuation factor
		if(q==0) continue; // terminates in a straight line?
		l = l/q; // q is in range (0,1] where 0 is a 180 degree turn, and 1 is a 0 degree turn (straight line)
		dx = pnx[i2]*l;
		dy = pny[i2]*l;
		px = rp[i2].x*scale - dx;
		py = rp[i2].y*scale - dy;
		op[2].x = -zx + cs*px - sn*py;
		op[2].y = -zy + sn*px + cs*py;
		px = rp[i2].x*scale + dx;
		py = rp[i2].y*scale + dy;
		op[3].x = -zx + cs*px - sn*py;
		op[3].y = -zy + sn*px + cs*py;

		DrawFilledQuad(op, foregroundcolor);
	}
}

void OriginalRasterizer::DrawRectangles( int num_rectangles, RenderRectangle * rr, double rotation, double scale)
{
	if(!bound_bmp || !rr) return;
	RenderPoint rp[4];
	int i;
	int n;
	double dx,dy,sn,cs;
	double c_w, c_h;

	// draw all the rectangles backgrounds
	cs = cos( rotation );
	sn = sin( rotation );
	for(i=0;i<num_rectangles;i++)
	{
		for(n=0;n<4;n++)
		{
			if(n==1 || n==2) c_w = -1; else c_w = 1;
			if(n<2) c_h = -1; else c_h = 1;
			dx = cs * (rr[i].w/2+shadow_thickness) * c_w - sn * (rr[i].h/2+shadow_thickness) * c_h;
			dy = sn * (rr[i].w/2+shadow_thickness) * c_w + cs * (rr[i].h/2+shadow_thickness) * c_h;
			rp[n].x = rr[i].x+dx-zx;
			rp[n].y = rr[i].y+dy-zy;
		}
		DrawFilledQuad(rp, backgroundcolor);
	}
	// draw all the rectangles
	for(i=0;i<num_rectangles;i++)
	{
		for(n=0;n<4;n++)
		{
			if(n==1 || n==2) c_w = -1; else c_w = 1;
			if(n<2) c_h = -1; else c_h = 1;
			dx = cs * rr[i].w/2 * c_w - sn * rr[i].h/2 * c_h;
			dy = sn * rr[i].w/2 * c_w + cs * rr[i].h/2 * c_h;
			rp[n].x = rr[i].x+dx-zx;
			rp[n].y = rr[i].y+dy-zy;
		}
		DrawFilledQuad(rp, foregroundcolor);
	}
}

void OriginalRasterizer::FlushChanges()
{ // some advanced rasterizers might use this, but not I.

}


void OriginalRasterizer::DrawFilledQuad( RenderPoint * rp, Color c )
{
	if(!bound_bmp || !rp) return;
	RenderPoint lp[4];
	{ // rotate points to get the topmost in [0], and put the rightmost point of 1,3 in [1]
		int r = 0, n;
		if(rp[0].y<= rp[1].y && rp[0].y<= rp[2].y && rp[0].y<= rp[3].y)
		{
			r=0;
		}
		else if(rp[1].y<= rp[0].y && rp[1].y<= rp[2].y && rp[1].y<= rp[3].y)
		{
			r=1;
		}
		else if(rp[2].y<= rp[0].y && rp[2].y<= rp[1].y && rp[2].y<= rp[3].y)
		{
			r=2;
		}
		else
		{
			r=3;
		}
		for(n=0;n<4;n++) lp[n] = rp[(n+r)&3];

		if(lp[2].y<=lp[0].y) return;
	}
	int bw = bound_bmp->wid();
	int bh = bound_bmp->hei();
	Color * bits = (Color *)bound_bmp->bits();
	double ly,py,y,xl,xr,dxl,dxr;
	int ipx,ipy, ipx2;
	bool ssl, sss, ssr;
	y=lp[0].y;
	ly=lp[1].y;
	if(ly<lp[2].y) ly=lp[2].y;
	if(ly<lp[3].y) ly=lp[3].y;
	xl=xr=lp[0].x;
	dxl=dxr=0;
	ssl=sss=ssr=false;
	py = floor(y);
	if(py<y) py=py+1;
	if(lp[0].y!=lp[3].y) dxl = (lp[3].x-lp[0].x)/(lp[3].y-lp[0].y);
	if(lp[0].y!=lp[1].y) dxr = (lp[1].x-lp[0].x)/(lp[1].y-lp[0].y);
	while(py<=ly)
	{
		xl += (py-y)*dxl;
		xr += (py-y)*dxr;
		y=py;
		if(!ssl) // second span left
		{
			if(py>lp[3].y) { 
				ssl=true; 
				xl = lp[3].x;
				dxl=0;
				if(lp[3].y!=lp[2].y) dxl = (lp[3].x-lp[2].x)/(lp[3].y-lp[2].y);
				xl += (py-lp[3].y)*dxl;
			}
		}
		if(!ssr) // second span right
		{
			if(py>lp[1].y) { 
				ssr=true; 
				xr = lp[1].x;
				dxr=0;
				if(lp[1].y!=lp[2].y) dxr = (lp[1].x-lp[2].x)/(lp[1].y-lp[2].y);
				xr += (py-lp[1].y)*dxr;
			}
		}
		if(!sss) // second second span :P
		{
			if(py>lp[2].y)
			{
				sss=true;
				if(ssl)
				{
					xl = lp[2].x;
					dxl=0;
					if(lp[1].y!=lp[2].y) dxl = (lp[1].x-lp[2].x)/(lp[1].y-lp[2].y);
					xl += (py-lp[2].y)*dxl;				
				}
				else if(ssr)
				{
					xr = lp[2].x;
					dxr=0;
					if(lp[3].y!=lp[2].y) dxr = (lp[3].x-lp[2].x)/(lp[3].y-lp[2].y);
					xr += (py-lp[2].y)*dxr;	
				}
				else
				{ // haha, malformed quad. it'll wind up looking like a triangle.
				}
			}
		}
		// draw span! xl to xr;
		// rough cut; this needs to be better antialiased.
		ipy=py;
		if(ipy>=0 && ipy<bh)
		{
			if(xl<xr) { ipx = (int)xl;	ipx2 = (int)xr; }
			else { ipx = (int)xr;	ipx2 = (int)xl; }
			if(ipx<0) ipx=0;
			if(ipx2>=bw) ipx2=bw-1;
			for(; ipx<=ipx2; ipx++)
			{
				bits[ipx+ipy*bw] = c;
			}

		}

		py+=1.0;
	}
}